from sanic_session.base import BaseSessionInterface
from sanic_session.utils import ExpiringDict


class InMemorySessionInterface(BaseSessionInterface):
    def __init__(
            self, domain: str = None, expiry: int = 2592000,
            httponly: bool = True, cookie_name: str = 'session',
            prefix: str = 'session:',
            sessioncookie: bool = False, samesite: str = None,
            session_name='session'):

        super().__init__(
            expiry=expiry,
            prefix=prefix,
            cookie_name=cookie_name,
            domain=domain,
            httponly=httponly,
            sessioncookie=sessioncookie,
            samesite=samesite,
            session_name=session_name,
        )
        self.session_store = ExpiringDict()

    async def _get_value(self, prefix, sid):
        return self.session_store.get(self.prefix + sid)

    async def _delete_key(self, key):
        if key in self.session_store:
            self.session_store.delete(key)

    async def _set_value(self, key, data):
        self.session_store.set(
            key, data,
            self.expiry
        )



from .memcache import MemcacheSessionInterface
from .redis import RedisSessionInterface
from .memory import InMemorySessionInterface
from .mongodb import MongoDBSessionInterface
from .aioredis import AIORedisSessionInterface
from .base import BaseSessionInterface

__all__ = ('MemcacheSessionInterface', 'RedisSessionInterface',
           'InMemorySessionInterface', 'MongoDBSessionInterface',
           'AIORedisSessionInterface', 'Session')


class Session:

    def __init__(self, app=None, interface=None):
        if app:
            self.init_app(app, interface=interface)

    def init_app(self, app, interface=None):
        if interface is None:
            interface = InMemorySessionInterface()

        self.interface = interface
        if not hasattr(app, 'extensions'):
            app.extensions = {}

        app.extensions[interface.session_name] = self  # session_name defaults to 'session'

        # @app.middleware('request')
        async def add_session_to_request(request):
            """Before each request initialize a session
            using the client's request."""
            await self.interface.open(request)

        # @app.middleware('response')
        async def save_session(request, response):
            """After each request save the session, pass
            the response to set client cookies.
            """
            await self.interface.save(request, response)

        app.request_middleware.appendleft(add_session_to_request)
        app.response_middleware.append(save_session)


from sanic_session.base import BaseSessionInterface

try:
    import aioredis
except ImportError:
    aioredis = None


class AIORedisSessionInterface(BaseSessionInterface):
    def __init__(
            self, redis,
            domain: str = None, expiry: int = 2592000,
            httponly: bool = True, cookie_name: str = 'session',
            prefix: str = 'session:',
            sessioncookie: bool = False, samesite: str = None,
            session_name: str = 'session'):
        """Initializes a session interface backed by Redis.

        Args:
            redis (Callable):
                aioredis connection or connection pool instance.
            domain (str, optional):
                Optional domain which will be attached to the cookie.
            expiry (int, optional):
                Seconds until the session should expire.
            httponly (bool, optional):
                Adds the `httponly` flag to the session cookie.
            cookie_name (str, optional):
                Name used for the client cookie.
            prefix (str, optional):
                Memcache keys will take the format of `prefix+session_id`;
                specify the prefix here.
            sessioncookie (bool, optional):
                Specifies if the sent cookie should be a 'session cookie', i.e
                no Expires or Max-age headers are included. Expiry is still
                fully tracked on the server side. Default setting is False.
            samesite (str, optional):
                Will prevent the cookie from being sent by the browser to the target
                site in all cross-site browsing context, even when following a regular link.
                One of ('lax', 'strict')
                Default: None
            session_name (str, optional):
                Name of the session that will be accessible through the request.
                e.g. If ``session_name`` is ``alt_session``, it should be accessed like that: ``request['alt_session']``
                e.g. And if ``session_name`` is left to default, it should be accessed like that: ``request['session']``
                Default: 'session'
        """
        if aioredis is None:
            raise RuntimeError(
                "Please install aioredis: pip install sanic_session[aioredis]")

        self.redis = redis

        super().__init__(
            expiry=expiry,
            prefix=prefix,
            cookie_name=cookie_name,
            domain=domain,
            httponly=httponly,
            sessioncookie=sessioncookie,
            samesite=samesite,
            session_name=session_name,
        )

    async def _get_value(self, prefix, sid):
        return await self.redis.get(self.prefix + sid)

    async def _delete_key(self, key):
        await self.redis.delete(key)

    async def _set_value(self, key, data):
        await self.redis.setex(key, self.expiry, data)



import time
import datetime
import abc
import ujson
import uuid
from sanic_session.utils import CallbackDict


class SessionDict(CallbackDict):
    def __init__(self, initial=None, sid=None):
        def on_update(self):
            self.modified = True

        super().__init__(initial, on_update)

        self.sid = sid
        self.modified = False


class BaseSessionInterface(metaclass=abc.ABCMeta):
    # this flag show does this Interface need request/responce middleware hooks

    def __init__(self, expiry, prefix, cookie_name, domain, httponly, sessioncookie, samesite, session_name):
        self.expiry = expiry
        self.prefix = prefix
        self.cookie_name = cookie_name
        self.domain = domain
        self.httponly = httponly
        self.sessioncookie = sessioncookie
        self.samesite = samesite
        self.session_name = session_name

    def _delete_cookie(self, request, response):
        response.cookies[self.cookie_name] = request[self.session_name].sid

        # We set expires/max-age even for session cookies to force expiration
        response.cookies[self.cookie_name]['expires'] = datetime.datetime.utcnow()
        response.cookies[self.cookie_name]['max-age'] = 0

    @staticmethod
    def _calculate_expires(expiry):
        expires = time.time() + expiry
        return datetime.datetime.fromtimestamp(expires)

    def _set_cookie_props(self, request, response):
        response.cookies[self.cookie_name] = request[self.session_name].sid
        response.cookies[self.cookie_name]['httponly'] = self.httponly

        # Set expires and max-age unless we are using session cookies
        if not self.sessioncookie:
            response.cookies[self.cookie_name]['expires'] = self._calculate_expires(self.expiry)
            response.cookies[self.cookie_name]['max-age'] = self.expiry

        if self.domain:
            response.cookies[self.cookie_name]['domain'] = self.domain

        if self.samesite is not None:
            response.cookies[self.cookie_name]['samesite'] = self.samesite

    @abc.abstractmethod
    async def _get_value(self, prefix: str, sid: str):
        '''
        Get value from datastore. Specific implementation for each datastore.

        Args:
            prefix:
                A prefix for the key, useful to namespace keys.
            sid:
                a uuid in hex string
        '''
        raise NotImplementedError

    @abc.abstractmethod
    async def _delete_key(self, prefix: str, key: str):
        '''Delete key from datastore'''
        raise NotImplementedError

    @abc.abstractmethod
    async def _set_value(self, key: str, data: SessionDict):
        '''Set value for datastore'''
        raise NotImplementedError

    async def open(self, request) -> SessionDict:
        """
        Opens a session onto the request. Restores the client's session
        from the datastore if one exists.The session data will be available on
        `request.session`.


        Args:
            request (sanic.request.Request):
                The request, which a sessionwill be opened onto.

        Returns:
            SessionDict:
                the client's session data,
                attached as well to `request.session`.
        """
        sid = request.cookies.get(self.cookie_name)

        if not sid:
            sid = uuid.uuid4().hex
            session_dict = SessionDict(sid=sid)
        else:
            val = await self._get_value(self.prefix, sid)

            if val is not None:
                data = ujson.loads(val)
                session_dict = SessionDict(data, sid=sid)
            else:
                session_dict = SessionDict(sid=sid)

        # attach the session data to the request, return it for convenience
        request[self.session_name] = session_dict
        return session_dict

    async def save(self, request, response) -> None:
        """Saves the session to the datastore.

        Args:
            request (sanic.request.Request):
                The sanic request which has an attached session.
            response (sanic.response.Response):
                The Sanic response. Cookies with the appropriate expiration
                will be added onto this response.

        Returns:
            None
        """
        if 'session' not in request:
            return

        key = (self.prefix + request[self.session_name].sid)
        if not request[self.session_name]:
            await self._delete_key(key)

            if request[self.session_name].modified:
                self._delete_cookie(request, response)
            return

        val = ujson.dumps(dict(request[self.session_name]))
        await self._set_value(key, val)
        self._set_cookie_props(request, response)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

# Set requirements here
requirements = ('sanic', 'ujson')

extras_require = {
    'aioredis': ['aioredis>=1.0.0'],
    'redis': ['asyncio_redis'],
    'mongo': ['sanic_motor', 'pymongo'],
    'aiomcache': ['aiomcache>=0.5.2'],
}


setup(
    name='sanic_session',
    version='0.5.1',
    description=(
        'Provides server-backed sessions for Sanic '
        'using Redis, Memcache and more.'
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/subyraman/sanic_session',
    author='Suby Raman, Mikhail Kashkin',
    author_email='m@xen.ru',
    license='MIT',
    packages=['sanic_session'],
    # Kludge: Specifying requirements for setup and install works around
    # problem with easyinstall finding sanic_motor instead of sanic.
    # See similar problem:
    #   https://stackoverflow.com/questions/27497470/setuptools-finds-wrong-package-during-install
    #   https://github.com/numpy/numpy/issues/2434
    setup_requires=requirements,
    install_requires=requirements,
    extras_require=extras_require,
    zip_safe=False,
    keywords=['sessions', 'sanic', 'redis', 'memcache'],
    classifiers=[
        'Framework :: AsyncIO',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Internet :: WWW/HTTP :: Session',
    ]
)



from datetime import datetime, timedelta
from sanic_session.base import BaseSessionInterface
import warnings

try:
    from sanic_motor import BaseModel

    class _SessionModel(BaseModel):
        """Collection for session storing.

        Collection name (default session)

        Fields:
            sid
            expiry
            data:
                User's session data
        """
        pass

except ImportError:  # pragma: no cover
    _SessionModel = None


class MongoDBSessionInterface(BaseSessionInterface):
    def __init__(
            self, app, coll: str = 'session',
            domain: str = None,
            expiry: int = 30 * 24 * 60 * 60,
            httponly: bool = True,
            cookie_name: str = 'session',
            sessioncookie: bool = False,
            samesite: str = None,
            session_name: str = 'session'):
        """Initializes the interface for storing client sessions in MongoDB.

        Args:
            app (sanic.Sanic):
                Sanic instance to register listener('after_server_start')
            coll (str, optional):
                MongoDB collection name for session
            domain (str, optional):
                Optional domain which will be attached to the cookie.
            expiry (int, optional):
                Seconds until the session should expire.
            httponly (bool, optional):
                Adds the `httponly` flag to the session cookie.
            cookie_name (str, optional):
                Name used for the client cookie.
            sessioncookie (bool, optional):
                Specifies if the sent cookie should be a 'session cookie', i.e
                no Expires or Max-age headers are included. Expiry is still
                fully tracked on the server side. Default setting is False.
            samesite (str, optional):
                Will prevent the cookie from being sent by the browser to
                the target site in all cross-site browsing context, even when
                following a regular link.
                One of ('lax', 'strict')
                Default: None
            session_name (str, optional):
                Name of the session that will be accessible through the
                request.
                e.g. If ``session_name`` is ``alt_session``, it should be
                accessed like that: ``request['alt_session']``
                e.g. And if ``session_name`` is left to default, it should be
                accessed like that: ``request['session']``
                Default: 'session'
        """
        if _SessionModel is None:
            raise RuntimeError("Please install Mongo dependencies: "
                               "pip install sanic_session[mongo]")

        # prefix not needed for mongodb as mongodb uses uuid4 natively
        prefix = ''

        if httponly is not True:
            warnings.warn('''
                httponly default arg has changed.
                To spare you some debugging time, httponly is currently
                hardcoded as True. This message will be removed with the
                next release. And ``httponly`` will no longer be hardcoded
            ''', DeprecationWarning)

        super().__init__(
            expiry=expiry,
            prefix=prefix,
            cookie_name=cookie_name,
            domain=domain,
            # I'm gonna leave this as True because changing it might
            # be hazardous. But this should be changed to __init__'s
            # httponly kwarg instead of being hardcoded
            httponly=True,
            sessioncookie=sessioncookie,
            samesite=samesite,
            session_name=session_name,
        )

        # set collection name
        _SessionModel.__coll__ = coll

        @app.listener('after_server_start')
        async def apply_session_indexes(app, loop):
            """Create indexes in session collection
            if doesn't exist.

            Indexes:
                sid:
                    For faster lookup.
                expiry:
                    For document expiration.
            """
            await _SessionModel.create_index('sid')
            await _SessionModel.create_index('expiry', expireAfterSeconds=0)

    async def _get_value(self, prefix, key):
        value = await _SessionModel.find_one({'sid': key}, as_raw=True)
        return value['data'] if value else None

    async def _delete_key(self, key):
        await _SessionModel.delete_one({'sid': key})

    async def _set_value(self, key, data):
        expiry = datetime.utcnow() + timedelta(seconds=self.expiry)
        await _SessionModel.replace_one(
            {'sid': key},
            {
                'sid': key,
                'expiry': expiry,
                'data': data
            },
            upsert=True
        )
